
'''
This file is automatically generated; Do not edit it. :)
'''
VERSION_INFO = {
    'final': True,
    'version': '1.29.2.post1',
    'branch_nick': 'hotfix-1.29.2.post1',
    'revision_id': 'dcc30a084b9663708f572ddb02999597c08cb9ba',
    'revno': 11929
}
